/*
 * IntervalTimer.cpp
 *
 *  Created on: Aug 28, 2015
 *      Author: eric
 */
#include <time.h>
#include <string.h>
#include <errno.h>
#include <fastec/interval_timer.h>

#include <fastec/port.h>

namespace Fastec
{

IntervalTimer::IntervalTimer() :
    lastMeasurement(0.0)
{
    ts_start.tv_sec = 0;
    ts_start.tv_nsec = 0;
}

IntervalTimer::~IntervalTimer()
{

}


void IntervalTimer::start()
{
    // Capture start time
    if (fi_clock_gettime(CLOCK_MONOTONIC, &ts_start) != 0)
    {
	FI_OS_ERROR(errno, "fi_clock_gettime()");
    }
}

void IntervalTimer::stop()
{
    struct timespec ts_end;

    // All finished; capture stats
    if ((ts_start.tv_sec != 0) || (ts_start.tv_nsec != 0))
    {
        if (fi_clock_gettime(CLOCK_MONOTONIC, &ts_end) != 0)
        {
	    FI_RUNTIME_ERROR("Timer stop() without prior start()");
        }
        lastMeasurement = ((double) ts_end.tv_sec + (double) ts_end.tv_nsec / 1.0e9)
                          - ((double) ts_start.tv_sec + (double) ts_start.tv_nsec / 1.0e9);
        if (lastMeasurement < 0.0)
        {
            // Should be impossible
	    FI_LOGIC_ERROR("Measured negative time interval");
        }
        accumulate(lastMeasurement);
        ts_start.tv_sec = 0;
        ts_start.tv_nsec = 0;
    }
}

void IntervalTimer::snap()
{

    if ((ts_start.tv_sec == 0) && (ts_start.tv_nsec == 0))
    {
        // First snap; just record the time
        if (fi_clock_gettime(CLOCK_MONOTONIC, &ts_start) != 0)
        {
	    FI_OS_ERROR(errno, "fi_clock_gettime()");
        }
    }
    else
    {
        struct timespec ts_now;

        if (fi_clock_gettime(CLOCK_MONOTONIC, &ts_now) != 0)
        {
	    FI_OS_ERROR(errno, "fi_clock_gettime()");
        }

        lastMeasurement = ((double) ts_now.tv_sec + (double) ts_now.tv_nsec / 1.0e9)
                          - ((double) ts_start.tv_sec + (double) ts_start.tv_nsec / 1.0e9);
        if (lastMeasurement < 0.0)
        {
            // Should be impossible
	    FI_LOGIC_ERROR("Measured negative time interval");
        }
        accumulate(lastMeasurement);
        memcpy((void*)(&ts_start), (const void*)(&ts_now), sizeof(struct timespec));
    }
}

void IntervalTimer::reset()
{
    lastMeasurement = 0.0;
    ts_start.tv_sec = 0;
    ts_start.tv_nsec = 0;
    Statistics::reset();
}

void IntervalTimer::dump(FILE* fd)
{
    switch (Statistics::dump(fd))
    {
    case 9:
        fprintf(fd, " nSec\n");
        break;
    case 6:
        fprintf(fd, " uSec\n");
        break;
    case 3:
        fprintf(fd, " mSec\n");
        break;
    case 0:
        fprintf(fd, " seconds\n");
        break;
    default:
        fprintf(fd, " seconds(?)\n");
        break;
    }
    dumpHist(fd);
}

void IntervalTimer::dump(FILE* fd, const char* tagStr)
{
    fprintf(fd, "%32s: ", tagStr);
    dump(fd);
}

} /* namespace Fastec */
