
#include <fastec/statistics.h>

#ifdef UNIT_TEST
#include <stdlib.h>
#include <stdio.h>
#endif /* UNIT_TEST */

#ifdef _WIN32
#undef max
#undef min
#endif

namespace Fastec
{

Statistics::Statistics() :
    max(-DBL_MAX),
    min(DBL_MAX),
    k(0),
    Mk(0.0),
    Sk(0.0),
    low(-DBL_MAX),
    high(DBL_MAX),
    nBins(0),
    hist(0)
{
    mutex = new std::mutex();
}

Statistics::~Statistics()
{
    max = -DBL_MAX;
    min = DBL_MAX;
    mutex->lock();
    if (hist != NULL)
    {
        long* h = hist;
        hist = NULL;
        mutex->unlock();
        delete h;
    }
    else
    {
        mutex->unlock();
    }

    delete mutex;
}

void Statistics::reset(void)
{
    mutex->lock();
    max = -DBL_MAX;
    min = DBL_MAX;
    k = 0;
    if (hist != NULL)
    {
        for (unsigned i = 0; i < nBins + 2; i++)
        {
            hist[i] = 0;
        }
    }
    mutex->unlock();
}

void Statistics::accumulate(double x)
{
    double Mnext;
    double Snext;

    mutex->lock();

    /* Knuth, Volume 2 2nd. Edition, page 216 [Welford 1962] */
    k++;
    if (k == 1)
    {
        /* First sample initializes the recurrance */
        Sk = 0.0;
        Mk = x;
    }
    else
    {
        /* k'th sample, k >= 2 */
        Mnext = Mk + ((x - Mk) / this->k);
        Snext = Sk + ((x - Mk) * (x - Mnext));
        Mk = Mnext;
        Sk = Snext;
    }

    if (x > max)
        max = x;
    if (x < min)
        min = x;

    if (hist != NULL)
    {
        if (x < low)
            hist[0]++;
        else if (x > high)
        {
            hist[nBins+1]++;
        }
        else if (x == high)
        {
            hist[nBins]++;
        }
        else
        {
            unsigned index = (unsigned)(((x - low) / (high - low)) * (double)(nBins - 1)) + 1;
            hist[index]++;
        }
    }
    mutex->unlock();
    return;
}

double Statistics::getMean(void)
{
    double val;

    mutex->lock();
    val = (k == 0) ? 0.0 : Mk;
    mutex->unlock();
    return val;
}

double Statistics::getMax(void)
{
    return max;
}

double Statistics::getMin(void)
{
    return min;
}

long Statistics::getNSamples(void)
{
    return k;
}

void Statistics::startHistogram(double l, double h, unsigned b)
{
    mutex->lock();
    if (hist != NULL)
        delete hist;
    hist = new long[b + 2];
    low = l;
    high = h;
    for (unsigned i=0; i<(b + 2); i++)
        hist[i] = 0;
    nBins = b;
    mutex->unlock();
}

int Statistics::dump(FILE* fd)
{
    int exp = 0;
    double scale = 1.0;

    static struct
    {
        double lim;
        double scale;
        int exp;
    } table[] =
    {
        { 1.0e9,    0.001, -3},
        { 1.0e6,    1.0,  0},
        { 1.0e3,    1.0,  0},
        { 1.0,   1000.0,  3},
        { 1.0e-9,   1.0e6,  6},
        { 1.0e-12,  1.0e9,  9},
    };

    mutex->lock();
    if (k == 0)
    {
        fprintf(fd, "min/mean/max for          0:     -.--- /     -.--- /     -.---");

    }
    else
    {
        for (unsigned i=0; i<sizeof(table)/sizeof(table[0]); i++)
        {
            if (max > table[i].lim)
            {
                scale = table[i].scale;
                exp = table[i].exp;
                break;
            }
        }

        fprintf(fd,
                "min/mean/max for %10ld: %9.03f / %9.03f / %9.3f",
                k,
                min * scale,
                ((k == 0) ? 0.0 : Mk) * scale,
                max * scale);
    }
    mutex->unlock();
    return exp;
}

int Statistics::dump(FILE* fd, const char* tagStr)
{
    fprintf(fd, "%32s: ", tagStr);
    return dump(fd);
}

void Statistics::dumpHist(FILE* fd)
{
    mutex->lock();
    if (hist != NULL)
    {
        fprintf(fd, "\n");
        for (unsigned i = 0; i < nBins + 2; i++)
        {
            if (i == 0)
            {
                fprintf(fd, "  < %f %ld\n", low, hist[0]);
            }
            else if (i == (nBins + 1))
            {
                fprintf(fd, "  > %f %ld\n", high, hist[nBins-1]);
            }
            else
            {
                fprintf(fd, "    %f - %f %ld\n",
                        low + (i - 1) * (high - low) / (double)nBins,
                        low + i * (high - low) / (double)nBins,
                        hist[i]);
            }
        }
    }
    mutex->unlock();
}

#ifdef UNIT_TEST
int main(int argc, char* argv[])
{
    int i;
    Statistics stats;

    printf("Initially: %ld samples with mean == %9.5f\n", stats.getNSamples(), stats.getMean());
    stats.accumulate((double)1.0);
    printf("        1: %ld samples with mean == %9.5f\n", stats.getNSamples(), stats.getMean());
    stats.accumulate((double)2.0);
    stats.accumulate((double)3.0);
    printf("        3: %ld samples with mean == %9.5f\n", stats.getNSamples(), stats.getMean());

    printf("Reset\n");
    stats.reset();
    printf("Initially: %ld samples with mean == %9.5f\n", stats.getNSamples(), stats.getMean());

    for (i=0; i<10000; i++)
    {
        stats.accumulate((double)rand() / (double)RAND_MAX);
    }
    printf("     rand: %ld samples with mean == %9.5f\n", stats.getNSamples(), stats.getMean());
    printf("     rand: min == %9.5f, max == %9.5f\n", stats.getMin(), stats.getMax());

    printf("Reset\n");
    stats.reset();
    for (i=0; i<123456; i++)
    {
        stats.accumulate(100.0 * ((double)rand() / (double)RAND_MAX) - 50.0);
    }
    printf(" neg rand: %ld samples with mean == %9.5f\n", stats.getNSamples(), stats.getMean());
    printf(" neg rand: min == %9.5f, max == %9.5f\n", stats.getMin(), stats.getMax());

    return(0);
}

#endif /* UNIT_TEST */

}
/* end namespace Fastec */
