#include <fastec/port.h>
#include <stdio.h>

#ifdef __APPLE__
int fi_clock_gettime(clockid_t clk_id, struct timespec *tp)
{
    int retval = 0;
    struct timeval tv;

    static double conversion = -1.0;

    tv.tv_sec = 0;
    tv.tv_usec = 0;

    if (clk_id == CLOCK_REALTIME)
    {
        if ((retval = gettimeofday(&tv, NULL)) != 0)
        {
            return retval;
        }
        tp->tv_sec = tv.tv_sec;
        tp->tv_nsec = tv.tv_usec * 1000;

    }
    else if (clk_id == CLOCK_MONOTONIC)
    {
        if (conversion < 0.0)
        {
            // first time through; get Mach time info from
            // kernel.
            mach_timebase_info_data_t info;
            kern_return_t err;
            if ((err = mach_timebase_info(&info)) == 0)
            {
                if ((info.numer == 1) && (info.denom == 1))
                {
                    conversion = 0.0;
                }
                else
                {
                    conversion = 1e-6 * ( double ) info.numer / ( double ) info.denom;
                }
            }
        }

        uint64_t t = mach_absolute_time();
        if (conversion == 0.0)
        {
            tp->tv_sec = (uint32_t)(t / (uint64_t)1e9);
            tp->tv_nsec = (uint32_t)(t - (uint64_t)(tp->tv_sec) * (uint64_t)1e9);
        }
        else
        {
            double dt = (double)t * conversion;
            tp->tv_sec = (uint32_t)dt;
            tp->tv_nsec = (uint32_t)(1.0e9 * (dt - (double)tp->tv_sec));
        }
    }

    return 0;
}
#endif // __APPLE__

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdint.h>
#include "gettimeofday.h"

static double history[8];
static unsigned history_index = 0;
static unsigned init = 0;
static double frequency = 0.0;

int fi_clock_gettime(clockid_t clk_id, struct timespec *tp)
{

    if (clk_id == CLOCK_REALTIME)
    {
        int retval = 0;
        struct timeval tv;
        tv.tv_sec = 0;
        tv.tv_usec = 0;

        if ((retval = gettimeofday(&tv, NULL)) != 0)
        {
            return retval;
        }
        tp->tv_sec = tv.tv_sec;
        tp->tv_nsec = tv.tv_usec * 1000;

    }
    else if (clk_id == CLOCK_MONOTONIC)
    {
        LARGE_INTEGER time;

        if (init == 0)
        {
            LARGE_INTEGER f;
            QueryPerformanceFrequency(&f);
            frequency = (double)f.QuadPart;
            init = 1;
        }
        QueryPerformanceCounter(&time);
        double seconds = (double)time.QuadPart / frequency;

        history[history_index++] = seconds;
        if (history_index >= (sizeof(history) / sizeof(history[0])))
            history_index = 0;

        tp->tv_sec = (uint32_t)seconds; // Truncated seconds
        tp->tv_nsec = (uint32_t)((seconds - (double)tp->tv_sec) * 1.0e9);
    }

    return 0;
}

#endif // _WIN32
