/*!
 * @file lince5m_utils.c
 *
 * @brief This file contains frame_rate computation functions for the Lince5M sensor.
 *
 * Although it would be nice if all the sensor-specific functions were isolated
 * to the Camera Server, to keep the UI responsive we have some functions in the GUIs.
 *
 * @author Bonnie Packert
 * @date 2014-10-09
 *
 * Copyright (C) 2014 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */


//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------

#include <unistd.h>
#include <stdint.h>
#include <stdio.h>

// If building as camera embedded GUI.
#ifdef TS3EMBGUI
#include <wchar.h>
#define MWINCLUDECOLORS
#include "nano-X.h"
#include "ts3_system.h"
#endif

// For FasMotion or for building this library
#include <fastec/gige_msg.h>
#include <fastec/lince5m_param.h>
#include <fastec/lince5m_utils.h>

// If building as standalone utility.
#ifdef SAUTIL
extern int verbose;
#endif

/*!
 * @brief Convert user ROI to sensor ROI
 *
 * Convert user ROI coordinates to sensor coordinates. Generally this is a matter of multiplying
 * by the binning/subsampling factor, but if the factor is 1 we apply an offset so the user
 * is always working with the center 1920x1080 (8- or 10-bit) or 1280x1024 (12-bit).
 *
 * @param factor	The binning/subsampling factor.
 * @param startx	Starting x offset in the user's ROI.
 * @param starty	Starting y offset in the user's ROI.
 * @param width		The width of the user's ROI.
 * @param height	The height of the user's ROI.
 * @param bit_mode	The bit mode to use, which can affect the interpretation of the ROI.
 *			The need for this dates back to the original implementation with different
 *			maximum sizes for different bit modes.
 * @param rtn_startx	Return the starting x offset in sensor coordinates.
 * @param rtn_starty	Return the starting y offset in sensor coordinates.
 * @param rtn_width	Return the width in sensor coordinates.
 * @param rtn_height	Return the height in sensor coordinates.
 *
 * @return nonzero if we had to adjust the ROI to make it legal, zero if all is okay.
 */
int lince5m_user_to_sensor(
    unsigned factor,
    uint16_t startx,
    uint16_t starty,
    uint16_t width,
    uint16_t height,
    uint8_t  bit_mode,
    uint16_t *rtn_startx,
    uint16_t *rtn_starty,
    uint16_t *rtn_width,
    uint16_t *rtn_height)
{
    int any_changes = 0;

    *rtn_startx = startx * factor;
    *rtn_starty = starty * factor;
    *rtn_width  = width  * factor;
    *rtn_height = height * factor;

    // Now center the user-visible region within what we allow.
    // We originally had different maximum sizes for different bit modes, which required
    // some fudging of offsets to hide that fact.
    if ( factor == 1 )
    {
        unsigned offset;
        uint16_t max_width;
        uint16_t max_height;

        lince5m_max_roi_setting( bit_mode, &max_width, &max_height );

        offset = ( LINCE_SENSOR_WIDTH - max_width ) / 2;
        if (( *rtn_startx + *rtn_width + offset ) <= LINCE_SENSOR_WIDTH )
            *rtn_startx += offset;

        offset = ( LINCE_SENSOR_HEIGHT - max_height ) / 2;
        if (( *rtn_starty + *rtn_height + offset ) <= LINCE_SENSOR_HEIGHT )
            *rtn_starty += offset;
    }

    if ( *rtn_width > LINCE_SENSOR_WIDTH )
    {
        *rtn_width = LINCE_SENSOR_WIDTH;
        any_changes = 1;
    }
    if ( *rtn_height > LINCE_SENSOR_HEIGHT )
    {
        *rtn_height = LINCE_SENSOR_HEIGHT;
        any_changes = 1;
    }
    if ( *rtn_startx + *rtn_width > LINCE_SENSOR_WIDTH )
    {
        *rtn_startx = LINCE_SENSOR_WIDTH - *rtn_width;
        any_changes = 1;
    }
    if ( *rtn_starty + *rtn_height > LINCE_SENSOR_HEIGHT )
    {
        *rtn_starty = LINCE_SENSOR_HEIGHT - *rtn_height;
        any_changes = 1;
    }
#ifdef SAUTIL
    if ( any_changes && verbose )
        printf( "Changed ROI to (%u,%u)%ux%u\n", *rtn_startx, *rtn_starty, *rtn_width, *rtn_height);
#endif
    return any_changes;
}

/*!
 * @brief Get the readout dims we need to give the sensor to satisfy boundary constraints
 *
 * Sensor readout is always in Pixel Processing Windows, 16 pixels, but the column window
 * we set up in the sensor must be at multiples of 4 PPWs or 64 pixels. Then if we are
 * binning or subsampling, the readout is guaranteed to be in multiples of 16 PPWs, since
 * the worst-case binning factor is 4.
 *
 * @param factor	The binning/subsampling factor
 * @param sensor_startx	The first sensor pixel in the user's ROI, in sensor coordinates
 * @param sensor_width	The width of the user's ROI, in sensor coordinates
 * @param rtn_startx	Return the first column to read in sensor coordinates
 * @param rtn_width	Return the number of column to read in sensor coordinates
 *
 * @return nonzero if we had to adjust the ROI to make it legal, zero if all is okay.
 */
int lince5m_get_readout_dims(
    unsigned factor,
    uint16_t sensor_startx,
    uint16_t sensor_width,
    uint16_t *rtn_startx,
    uint16_t *rtn_width )
{
    int      any_changes = 0;
    uint16_t readout_startx;
    uint16_t readout_endx;
//  unsigned multiple;

    // Now we need to make sure the readout for each row is a multiple of 16 pixels even
    // after it has been subsampled. The startx we give the sensor must be a multiple of 64
    // to accommodate sensor restrictions.
//  multiple = factor * PIXELS_PER_BLOCK;

    // This is the next lower multiple of 64 below startx.
    readout_startx = ( sensor_startx / PIXELS_PER_BLOCK ) * PIXELS_PER_BLOCK;
    // This is the next multiple of 64 higher than the rightmost pixel.
    readout_endx = (( sensor_startx + sensor_width + PIXELS_PER_BLOCK - 1 ) / PIXELS_PER_BLOCK ) *
                   PIXELS_PER_BLOCK;

    // When we readout, the FPGA wants a multiple of 16 pixels and there must be at least 64 pixels
    // for the arithmetic in last crop area. So as long as we have 5 PPWs left after binning,
    // we should be okay.
    if ((unsigned)( readout_endx - readout_startx ) < ( factor * ( 5 * PIXELS_PER_PPW )))
    {
        if ( readout_startx >= PIXELS_PER_BLOCK )
            readout_startx -= PIXELS_PER_BLOCK;
        if ((unsigned)( readout_endx - readout_startx ) < ( factor * ( 5 * PIXELS_PER_PPW )))
        {
            if ( readout_endx < ( LINCE_SENSOR_WIDTH - PIXELS_PER_BLOCK ))
                readout_endx += PIXELS_PER_BLOCK;
        }
        if ((unsigned)( readout_endx - readout_startx ) < factor * ( 5 * PIXELS_PER_PPW ))
            any_changes = 1;
    }

    *rtn_startx = readout_startx;
    *rtn_width = readout_endx - readout_startx;
#ifdef SAUTIL
    if ( verbose )
    {
        printf( "Sensor=%u:%u - Readout=%u:%u\n", sensor_startx, sensor_width, *rtn_startx, *rtn_width );
    }
#endif

    return any_changes;
}

/*!
 * @brief Compute the maximum frame rate available for a given ROI.
 *
 * @param readout_width	    The number of columns being read out from the sensor.
 * @param readout_height    The number of rows being read out from the sensor.
 * @param sensopts	    Sensor options to get the framing mode
 *
 * @return The maximum number of frames per second for this size and framing mode.
 */
uint32_t lince5m_get_max_frame_rate(
    uint32_t readout_width,
    uint32_t readout_height,
    uint32_t sensopts )
{
    uint32_t t_row_cycles;

    t_row_cycles = lince5m_get_minimum_trow( readout_width, sensopts );
    return lince5m_get_max_frame_rate_trow( readout_width, readout_height, sensopts,
                                            t_row_cycles );
}

/*!
 * @brief Compute the maximum frame rate for a given size with a given t_row choice.
 *
 * Frame time is the maximum of t_exp and t_readout.
 *
 * The value for t_exp includes the sensor states B, E, and F (t_exp_B, t_exp_E, t_exp_F),
 * where t_exp_E is the exposure time, t_exp_B is N_cycles_B / f_sys, and
 * t_exp_F = N_cycles_F / fsys).
 *
 * t_exp = t_exp_B + t_exp_E + t_exp_F
 *
 * The values for N_cycles_B and N_cycles_F come from the configuration, but we need those
 * values before the sensor is configured, so we just pre-define them based on what was in
 * those .dat files. There are separate configuration files for the three t_row values we support.
 *
 * t_readout = t_row_cycles * rows + extra-cycles + t_wait_cycles + N_cycles_F.
 *
 * @param readout_width	    The number of columns being read out from the sensor.
 * @param readout_height    The number of rows being read out from the sensor.
 * @param sensopts	    Sensor options to get the framing mode
 * @param t_row_cycles	    The t_row value to use, which is not necessarily the lowest
 *			    possible value we can use for this readout width.
 *
 * @return The maximum number of frames per second
 */
uint32_t lince5m_get_max_frame_rate_trow(
    uint32_t readout_width,
    uint32_t readout_height,
    uint32_t sensopts,
    uint32_t t_row_cycles )
{
    uint32_t rows;
    uint32_t t_readout;
    uint32_t t_wait_cycles;
    uint32_t N_cycles_B, N_cycles_F;
    uint32_t K;
    uint32_t fps;
    uint32_t f_sys;
    int      use_fm6;

    if (( sensopts & GIGE_SENSOPT_INTERNAL_0 ) != 0 )
    {
        use_fm6 = 1;
        f_sys = FM6_F_SYS;
    }
    else
    {
        use_fm6 = 0;
        f_sys = FM1_F_SYS;
    }

    // t_row_min depends on the loaded configuration, which depends on the readout width,
    // but we don't add any HBlanking_cycles. So t_row_cycles=t_row_min.
    if ( t_row_cycles == TROW_52 )
    {
        N_cycles_B = FM1_N_CYCLES_B_52;
        N_cycles_F = FM1_N_CYCLES_F_52;
    }
    else if ( t_row_cycles == TROW_84 )
    {
        N_cycles_B = use_fm6 ? FM6_N_CYCLES_B_84 : FM1_N_CYCLES_B_84;
        N_cycles_F = use_fm6 ? FM6_N_CYCLES_F_84 : FM1_N_CYCLES_F_84;
    }
    else if ( t_row_cycles == TROW_124 )
    {
        N_cycles_B = use_fm6 ? FM6_N_CYCLES_B_124 : FM1_N_CYCLES_B_124;
        N_cycles_F = use_fm6 ? FM6_N_CYCLES_F_124 : FM1_N_CYCLES_F_124;
    }
    else if ( t_row_cycles == TROW_132 )
    {
        N_cycles_B = use_fm6 ? FM6_N_CYCLES_B_132 : FM1_N_CYCLES_B_132;
        N_cycles_F = use_fm6 ? FM6_N_CYCLES_F_132 : FM1_N_CYCLES_F_132;
    }
    else
    {
        t_row_cycles = TROW_165;    // Any invalid value defaults to 165.
        N_cycles_B = use_fm6 ? FM6_N_CYCLES_B_165 : FM1_N_CYCLES_B_165;
        N_cycles_F = use_fm6 ? FM6_N_CYCLES_F_165 : FM1_N_CYCLES_F_165;
    }

    // (24) K is ( N_cycles_B / t_row_cycles ) rounded up plus 3.
    K = (( N_cycles_B + t_row_cycles - 1 ) / t_row_cycles ) + 3;
    rows = readout_height + EXTRA_ROWS + K;

    // (28) Extra time for internal synchronization.
    // The +1 is from the AnaFocus frame rate spreadsheet, which disagrees with the data sheet.
    t_wait_cycles = 3 * t_row_cycles + 1;

    // This is the readout time in f_sys cycles.
    t_readout = ( t_row_cycles * rows ) + EXTRA_CYCLES + t_wait_cycles + N_cycles_F;

    // Now convert to frames per second. The readout time is in f_sys cycles so
    // we can just divide f_sys by t_readout and it will just round down, which
    // is what we want.
    fps = f_sys / t_readout;

// If running standalone utility.
#ifdef SAUTIL
    if ( verbose )
    {
        printf( "t_row_cycles=%u\n", t_row_cycles );
        printf( "K=%u\n", K );
        printf( "rows=%u (%u + %u + %u)\n", rows, readout_height, EXTRA_ROWS, K );
        printf( "Extra_cycles=%u\n", EXTRA_CYCLES );
        printf( "t_wait_cycles=%u\n", t_wait_cycles );
        printf( "N_cycles_B=%u\n", N_cycles_B );
        printf( "N_cycles_F=%u\n", N_cycles_F );
        printf( "%ux%u t_readout=%f=%u cycles, fps=%u\n", readout_width, readout_height,
                (double)t_readout / (double)(f_sys/1000), t_readout, fps );
    }
#endif

    return fps;
}

/*!
 * @brief Compute the maximum frame rate from GUI-visible parameters
 *
 * @param startx    The ROI start column, relative to what is visible to the GUI
 * @param starty    The ROI start row, relative to what is visible to the GUI
 * @param width     The ROI effective width
 * @param height    The ROI effective height
 * @param bit_mode  The bit depth option
 * @param sensopts  Sensor binning and subsampling options
 *
 * The bit_mode was needed originally because we limited the ROI size based on the bit mode,
 * but that no longer is true so the parameter could be retired.
 *
 * @return The maximum number of frames per second
 */
uint32_t lince5m_max_fps_setting(
    uint16_t startx,
    uint16_t starty,
    uint16_t width,
    uint16_t height,
    uint8_t  bit_mode,
    uint32_t sensopts )
{
    uint16_t sensor_x;
    uint16_t sensor_y;
    uint16_t sensor_width;
    uint16_t sensor_height;
    uint16_t readout_x;
    uint16_t readout_width;
    uint32_t fps;
    unsigned factor;

    // Get the binning/subsampling factor.
    factor = lince5m_binsubs_factor( sensopts );

    (void)lince5m_user_to_sensor( factor, startx, starty, width, height, bit_mode,
                                  &sensor_x, &sensor_y, &sensor_width, &sensor_height );
    lince5m_get_readout_dims( factor, sensor_x, sensor_width, &readout_x, &readout_width );

    fps = lince5m_get_max_frame_rate( readout_width / factor, sensor_height / factor, sensopts );

    // Return the integer maximum frame rate than can be supported with these parameters.
    return fps;
}

/*!
 * @brief GUI support function to get maximum size for a bit mode
 *
 * @param bit_mode      The bit depth option
 * @param rtn_width     Return maximum permitted width for bit_mode
 * @param rtn_height    Return maximum permitted height for bit_mode
 *
 * @return The maximum number of frames per second
 */
void lince5m_max_roi_setting( uint8_t bit_mode, uint16_t *rtn_width, uint16_t *rtn_height )
{
    if ( IS_12_BIT_MODE( bit_mode ))
    {
        *rtn_width  = LINCE_MAX_WIDTH_12BIT;
        *rtn_height = LINCE_MAX_HEIGHT_12BIT;
    }
    else if ( IS_10_BIT_MODE( bit_mode ))
    {
        *rtn_width  = LINCE_MAX_WIDTH_10BIT;
        *rtn_height = LINCE_MAX_HEIGHT_10BIT;
    }
    else /* IS_8_BIT_MODE( bit_mode ) */
    {
        *rtn_width = LINCE_REPORTED_MAX_WIDTH;
        *rtn_height = LINCE_REPORTED_MAX_HEIGHT;
    }
}

/*!
 * @brief Get the binning/subsampling multiplier.
 *
 * @param sensor_opts	Sensor options, which includes binning/subsampling choices
 *
 * @return the binning/subsampling factor
 */
unsigned lince5m_binsubs_factor( uint32_t sensopts )
{
    unsigned factor;

    factor = (( sensopts & GIGE_SENSOPT_BIN_MASK ) >> GIGE_SENSOPT_BIN_SHIFT ) *
             (( sensopts & GIGE_SENSOPT_SUBS_MASK ) >> GIGE_SENSOPT_SUBS_SHIFT );
    // Be safe. Be sane.
    if ( factor == 0 )
        factor = 1;

    return factor;
}

/*!
 * @brief Get the offset of the capturable area from the max capturable area.
 *
 * @param bit_mode	Bit mode we are capturing with
 * @param sensor_opts	Sensor options, which includes binning/subsampling choices
 * @param *rtn_ofs_x	On return, the "hidden" X offset due to bandwidth control ROI limits
 * @param *rtn_ofs_y	On return, the "hidden" Y offset due to bandwidth control ROI limits
 *
 * @return the binning/subsampling factor
 */
void lince5m_get_ct_offset(
    uint8_t  bit_mode,
    uint32_t sensor_opts,
    uint16_t *rtn_ofs_x,
    uint16_t *rtn_ofs_y )
{
    if ( IS_12_BIT_MODE( bit_mode ) && ( lince5m_binsubs_factor( sensor_opts ) == 1 ))
    {
        uint16_t max_width;
        uint16_t max_height;

        lince5m_max_roi_setting( bit_mode, &max_width, &max_height );

        *rtn_ofs_x = ( LINCE_REPORTED_MAX_WIDTH - max_width ) / 2;
        *rtn_ofs_y = ( LINCE_REPORTED_MAX_HEIGHT - max_height ) / 2;
    }
    else
    {
        *rtn_ofs_x = 0;
        *rtn_ofs_y = 0;
    }
}

/*
 * @brief Get the t_row value based on the sensor options and the readout width
 *
 * @param readout_width	The width of lines being read out of the sensor
 * @param sensor_opts	The sensor options, in particular, binning info.
 *
 * @return the t_row value to use.
 */
uint32_t lince5m_get_minimum_trow( uint32_t readout_width, uint32_t sensor_opts )
{
    uint32_t t_row_cycles;

    if (( readout_width <= TROW_52_MAX_WIDTH ) &&
            ((( sensor_opts & GIGE_SENSOPT_BIN_MASK ) >> GIGE_SENSOPT_BIN_SHIFT ) > 1 ))
    {
        t_row_cycles = TROW_52;
    }
    else if ( readout_width <= TROW_84_MAX_WIDTH )
    {
        t_row_cycles = TROW_84;
    }
    else if ( readout_width <= TROW_124_MAX_WIDTH )
    {
        t_row_cycles = TROW_124;
    }
    else if ( readout_width <= TROW_132_MAX_WIDTH )
    {
        t_row_cycles = TROW_132;
    }
    else
    {
        t_row_cycles = TROW_165;
    }

    return t_row_cycles;
}

/*
 * @brief Get the t_row value based on the sensor options, frame_rate, and user ROI
 *
 * @param x		User's ROI x offset, adjusted based on license restrictions
 * @param y		User's ROI y offset, adjusted based on license restrictions
 * @param width		User's ROI width
 * @param height	User's ROI height
 * @param bit_mode	User's bit mode, because it's still passed even though we don't need it
 * @param sensor_opts	The sensor options, in particular, binning info.
 * @param frame_rate	The frame rate
 *
 * The (x,y) offset is adjusted from the user-visible offset based on license restrictions
 * that hide the outer portions of the sensor.
 *
 * @return the t_row value to use.
 */
uint32_t lince5m_get_trow_from_user(
    uint32_t startx,
    uint32_t starty,
    uint32_t width,
    uint32_t height,
    uint8_t  bit_mode,
    uint32_t sensor_opts,
    uint32_t frame_rate )
{
    uint16_t sensor_x;
    uint16_t sensor_y;
    uint16_t sensor_width;
    uint16_t sensor_height;
    uint16_t readout_x;
    uint16_t readout_width;
    unsigned factor;
    unsigned t_row;

    // Get the binning/subsampling factor.
    factor = lince5m_binsubs_factor( sensor_opts );

    (void)lince5m_user_to_sensor( factor, startx, starty, width, height, bit_mode,
                                  &sensor_x, &sensor_y, &sensor_width, &sensor_height );
    lince5m_get_readout_dims( factor, sensor_x, sensor_width, &readout_x, &readout_width );

    t_row = lince5m_get_minimum_trow( readout_width, sensor_opts );

    // If the size allows us to use a reduced t_row value, see if our frame rate actually requires
    // that we use that or if instead we can get away the cleaner configuration for width <=1920.
    // Otherwise flag that we are using the fast-mode configuration.
    if (( t_row == TROW_84 ) || ( t_row == TROW_52 ))
    {
        uint32_t alt_limit;

        alt_limit = lince5m_get_max_frame_rate_trow( readout_width / factor, sensor_height / factor,
                    sensor_opts, TROW_124 );
        if ( frame_rate <= alt_limit )
            t_row = TROW_124;
    }
    return t_row;
}

