/*!
 * @file lssm.c
 *
 * @brief Local Stream Shared Memory interfaces
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

// For C code and Ubuntu 10.10 builds, we have to define __USE_UNIX98 to get the
// pthread_mutexattr_settype function and its values. This is not needed for our
// C++ code, and probably won't be needed for newer versions.
#define __USE_UNIX98
#include <pthread.h>
static pid_t lssm_gettid();

#include <fastec/lssm.h>

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Name of shared memory region
#define LOCALSTREAM_SHM_NAME "/cs_wsd_localstream"
#define LOCALSTREAM_SHM_SIZE 4096
#define LOCAL_STREAM_SHM_HEADER_RAM_LENGTH 0x800

#define BFSTATE_EMPTY   0   //!< Available to write an image
#define BFSTATE_WRITING 1   //!< Server is writing image
#define BFSTATE_FULL    2   //!< Available to be read
#define BFSTATE_READING 3   //!< Client is reading image

#define NUM_BUFS  2         //!< Number of image memory buffers

//--------------------------------------------------------------------
// Local variables
//--------------------------------------------------------------------

// --------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------

// --------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

//! Shared buffer descriptor.
typedef struct lssmbuf
{
    uint64_t im_addr;       //!< Address in system memory.
    uint32_t frame_size;    //!< Size of processed image in bytes.
    int      frame_num;     //!< Frame number (0 to framecount-1), negative illegal.
    uint64_t sequence;      //!< Sequence number for finding oldest frame.
    uint8_t  state;         //!< Buffer state, BFSTATE_...
} lssmbuf_t;

//! Opaque type always accessed through function interfaces.
typedef struct lssm
{
    pthread_mutex_t stream_lock;    //!< Process-shared mutex.
    pthread_cond_t server_wakeup;   //!< Condition variable for signaling server.
    pthread_cond_t client_wakeup;   //!< Condition variable for signaling client.

    uint32_t server_fault;          //!< Fault indicator from server; client clears (non-zero => fault)
    uint32_t client_fault;          //!< Fault indicator from client; server clears (non-zero => fault)
    uint32_t buf_size;              //!< Allocated size of image memory buffers
    uint64_t last_sequence;         //!< For finding oldest frame.
    int32_t mode;                   //!< Current local stram mode ()GIGE_LOCALSTREAM_*)

    // Metadata for each of the buffers.
    lssmbuf_t imagebuf[NUM_BUFS];

    //! Image header bytes for tif, bmp, and dng formats.
    uint8_t image_header[LOCAL_STREAM_SHM_HEADER_RAM_LENGTH];
    uint32_t header_length;   //!< Number of bytes in image_header to prepend to each frame.

} lssm_t;


/*!
 * @brief Take the shared memory mutex.
 *
 * @param p_shared_mem  Either client's or server's pointer to shared memory.
 *
 * @returns non-zero for error, 0 for success.
 */
static pid_t last_locker = 0;   // For debug messages.
int lssm_lock(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    rc = pthread_mutex_trylock(&p_lssm->stream_lock);
    if (rc != 0)
    {
        if (rc == EBUSY)
        {
            rc = pthread_mutex_lock(&p_lssm->stream_lock);
        }
    }

    // See if we got the lock.
    if (rc == 0)
    {
        last_locker = lssm_gettid();
    }
    else
    {
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));
    }

    return rc;
}

/*!
 * @brief Same as lssm_lock(), but with a timeout.
 */
int lssm_timedlock(void *p_shared_mem, struct timespec *tm_ptr)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
        syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    // Try locking the mutex to see if we get lucky.
    rc = pthread_mutex_trylock(&p_lssm->stream_lock);
    if (rc != 0)
    {
        // Is the mutex busy (i.e., already locked)?
        if (rc == EBUSY)
        {
            // Use the appropriate mechanism to wait.
            if (tm_ptr == NULL)
                rc = pthread_mutex_lock(&p_lssm->stream_lock);
            else
                rc = pthread_mutex_timedlock(&p_lssm->stream_lock, tm_ptr);
        }
    }

    // See if we got the lock.
    if (rc == 0)
    {
        // Got the lock. Save some debug info.
        last_locker = lssm_gettid();
    }
    else
    {
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));
    }

    return rc;
}

/*!
 * @brief Release the shared memory mutex.
 *
 * @param p_shared_mem  Either client's or server's pointer to shared memory.
 *
 * @returns non-zero for error, 0 for success.
 */
int lssm_unlock(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    last_locker = 0;
    rc = pthread_mutex_unlock(&p_lssm->stream_lock);
    if (rc < 0)
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));

    return rc;
}

/*!
 * @brief Wake up the server using the condition variable in shared memory.
 *
 * This needs to be a separate function from the lock/unlock to support proper
 * mutual exclusion when signalling:
 *
 *   lock
 *   set up memory content, e.g. add message to queue or return a buffer
 *   signal condition
 *   unlock
 *
 * @param p_shared_mem Server's pointer to shared memory.
 *
 * @returns non-zero for error, 0 for success.
 */
int lssm_server_signal_l(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    // Send the signal.
    rc = pthread_cond_signal(&p_lssm->server_wakeup);
    if (rc < 0)
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));

    return rc;
}

int lssm_server_clear_fault_l(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    // Clear the fault setting and send the signal.
    p_lssm->client_fault = 0;
    rc = pthread_cond_signal(&p_lssm->client_wakeup);
    if (rc != 0)
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));

    return rc;
}

int lssm_server_set_fault_l(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    syslog(LOG_ERR, "%s: Send LSSM Server fault indication to LSSM client", __FUNCTION__);
    p_lssm->server_fault = 1;
    rc = pthread_cond_signal(&p_lssm->client_wakeup);
    if (rc != 0)
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));

    return rc;
}

/*!
 * @brief Server call to wait for event or timeout
 *
 * @param p_shared_mem  Server's pointer to shared memory.
 * @param util_time     When to wake up
 *
 * @returns non-zero for error, 0 for success.
 */
int  lssm_server_wait_until_l(void *p_shared_mem, struct timespec *until_time)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int rc;

    // Sanity check.
    if (until_time == NULL)
    {
	syslog(LOG_ERR, "%s: wakeup time pointer is null!", __PRETTY_FUNCTION__);
	return EINVAL;
    }

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    rc = pthread_cond_timedwait(&p_lssm->server_wakeup, &p_lssm->stream_lock, until_time);
    if ((rc != 0) && (rc != ETIMEDOUT))
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(rc));

    return rc;
}

/*!
 * @brief Server call to wait for event or timeout
 *
 * @param p_shared_mem  Server's pointer to shared memory.
 * @param timeout_usecs Timeout in usecs
 *
 * @returns non-zero for error, 0 for success.
 */
int lssm_server_wait_l(void *p_shared_mem, uint64_t timeout_usecs)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    struct timespec ts;
    int             rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    if (timeout_usecs != 0)
    {
        clock_gettime(CLOCK_REALTIME, &ts);
        ts.tv_nsec += ((timeout_usecs % 1000000) * 1000);
        while (ts.tv_nsec >= 1000 * 1000 * 1000)
        {
            ts.tv_nsec -= 1000 * 1000 * 1000;
            ts.tv_sec += 1;
        }
        ts.tv_sec += (timeout_usecs / 1000000);

        rc = pthread_cond_timedwait(&p_lssm->server_wakeup, &p_lssm->stream_lock, &ts);
    }
    else
    {
        rc = pthread_cond_wait(&p_lssm->server_wakeup, &p_lssm->stream_lock);
    }

    if (rc < 0)
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));

    return rc;
}

/*!
 * @brief Get the number of buffer not in the EMPTY state.
 *
 * @param p_shared_mem  Pointer to shared memory.
 *
 * @returns The number of buffers with states other than EMPTY
 */
int lssm_non_empty_buffers_l(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int     i;
    int     rc = 0;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return 0;
    }

    for (i = 0; i < NUM_BUFS; ++i)
    {
        if (p_lssm->imagebuf[i].state != BFSTATE_EMPTY)
            rc++;
    }
    return rc;
}

int lssm_get_server_fault_l(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return 0;
    }

    return p_lssm->server_fault;
}

int lssm_get_client_fault_l(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return 0;
    }

    return p_lssm->client_fault;
}

int  lssm_get_mode(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return 0;
    }

    return (int)p_lssm->mode;
}

void lssm_set_mode(void *p_shared_mem, int mode)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm != NULL)
        p_lssm->mode = (int32_t)mode;
    else
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
}

/*!
 * @brief Get a buffer to write into
 *
 * When doing playback or singlestep, we don't need to preserve old frames
 * and can reclaim a full buffer and refill it with a newer frame. We take
 * the older one so the reader will still have a frame available while we
 * are busy filling the other one.
 * When doing download, we cannot drop any frames so we just have to return
 * an error.
 *
 * @param p_shared_mem      Pointer to shared memory.
 * @param overwrite_oldest  non-zero if we should overwrite an old buffer,
 *              zero to return an error if all buffers are full
 *              or being read.
 *
 * @returns buffer ID >= 0 of an available buffer to write into, negative for error
 */
int  lssm_get_write_buffer_l(void *p_shared_mem, int overwrite_oldest)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int i;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return -1;
    }

    for (i = 0; i < NUM_BUFS; ++i)
    {
        if (p_lssm->imagebuf[i].state == BFSTATE_EMPTY)
            break;
    }
    if (i != NUM_BUFS)
    {
        p_lssm->imagebuf[i].state = BFSTATE_WRITING;
        p_lssm->imagebuf[i].sequence = p_lssm->last_sequence++;
        return i;
    }

    if (overwrite_oldest)
    {
        int oldest = -1;
        uint32_t oldest_sequence = 0;

        syslog(LOG_DEBUG, "%s: OVERWRITE", __PRETTY_FUNCTION__);

        for (i = 0; i < NUM_BUFS; ++i)
        {
            if (p_lssm->imagebuf[i].state == BFSTATE_FULL)
            {
                if ((oldest == -1) ||
                    (oldest_sequence > p_lssm->imagebuf[i].sequence))
                {
                    oldest = i;
                    oldest_sequence = p_lssm->imagebuf[i].sequence;
                }
            }
        }
        if (oldest != -1)
        {
            p_lssm->imagebuf[oldest].state = BFSTATE_WRITING;
            p_lssm->imagebuf[oldest].sequence = p_lssm->last_sequence++;
            return oldest;
        }
    }

    syslog(LOG_ERR, "%s: No buffer found!", __PRETTY_FUNCTION__);
    return -1;
}

/*!
 * @brief Get the image memory address corresponding to the buffer
 *
 * @param p_shared_mem  Pointer to shared memory.
 * @param bufid         ID of the buffer whose address is to be returned.
 *
 * @returns image memory address or zero for error.
 */
uint64_t lssm_get_image_address(void *p_shared_mem, int bufid)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    uint64_t addr = 0;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return 0;
    }

    if ((bufid < 0) || (bufid >= NUM_BUFS))
    {
        syslog(LOG_ERR, "%s: bad bufid=%d", __FUNCTION__, bufid);
    }
    else
    {
        addr = p_lssm->imagebuf[bufid].im_addr;
    }
    return addr;
}

/*!
 * @brief Get the frame number corresponding to the buffer
 *
 * @param p_shared_mem  Pointer to shared memory.
 * @param bufid         ID of the buffer whose address is to be returned.
 *
 * @returns frame number or negative for error.
 */
int lssm_get_frame_num(void *p_shared_mem, int bufid)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int frame_num = -1;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return -1;
    }

    if ((bufid < 0) || (bufid >= NUM_BUFS))
    {
        syslog(LOG_ERR, "%s: bad bufid=%d", __FUNCTION__, bufid);
        return -1;
    }
    else
    {
        frame_num = p_lssm->imagebuf[bufid].frame_num;
    }
    return frame_num;
}

/*!
 * @brief Set the frame number corresponding to the buffer
 *
 * Does not enforce it, but this should only be done by the owner of the buffer.
 *
 * @param p_shared_mem  Pointer to shared memory.
 * @param bufid         ID of the buffer whose address is to be returned.
 * @param frame_num     The frame number to associate with the buffer.
 */
void lssm_set_frame_num(void *p_shared_mem, int bufid, int frame_num)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return;
    }

    if ((bufid < 0) || (bufid >= NUM_BUFS))
    {
        syslog(LOG_ERR, "%s: bad bufid=%d for frame #%u", __FUNCTION__, bufid, frame_num);
    }
    else
    {
        p_lssm->imagebuf[bufid].frame_num = frame_num;
    }
}

/*!
 * @brief Get image size corresponding to the buffer
 *
 * @param p_shared_mem      Pointer to shared memory.
 * @param bufid         ID of the buffer whose image size is to be returned.
 *
 * @returns Size in bytes or zero if error.
 */
uint32_t lssm_get_processed_size(void *p_shared_mem, int bufid)
{
    lssm_t   *p_lssm = (lssm_t *)p_shared_mem;
    uint32_t size = 0;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return 0;
    }

    if ((bufid < 0) || (bufid >= NUM_BUFS))
    {
        syslog(LOG_ERR, "%s: bad bufid=%d", __FUNCTION__, bufid);
    }
    else
    {
        size = p_lssm->imagebuf[bufid].frame_size;
    }
    return size;
}

/*!
 * @brief Set the frame size corresponding to the buffer
 *
 * Does not enforce it, but this should only be done by the owner of the buffer.
 *
 * @param p_shared_mem      Pointer to shared memory.
 * @param bufid         ID of the buffer whose address is to be returned.
 * @param processed_size    The size to associate with the buffer.
 *
 * @see lssm_get_processed_size
 */
void lssm_set_processed_size(void *p_shared_mem, int bufid, uint32_t processed_size)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return;
    }

    if ((bufid < 0) || (bufid >= NUM_BUFS))
    {
        syslog(LOG_ERR, "%s: bad bufid=%d for frame #%u", __FUNCTION__, bufid, processed_size);
    }
    else
    {
        p_lssm->imagebuf[bufid].frame_size = processed_size;
    }
}

/*!
 * @brief Get size (in bytes) of current image header
 *
 * @return Image header size, in bytes
 */
uint32_t lssm_get_image_header_size(void *p_shared_mem)
{
    // Sanity check.
    if (p_shared_mem == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return 0;
    }

    return ((lssm_t *)p_shared_mem)->header_length;
}

/*!
 * @brief Copy image header shared memory buffer into caller's image header buffer
 *
 * If caller's buffer is not of sufficient size, then no bytes are copied.
 *
 * @param p_shared_mem Pointer to opaque LSSM shared memory area
 * @param dst          Pointer to caller's buffer
 * @param src_size     Size (bytes) of caller's image header
 */
void lssm_copyin_image_header(void *p_shared_mem, uint8_t *dst, uint32_t dst_size)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return;
    }

    if ((p_lssm->header_length > 0) && (p_lssm->header_length <= dst_size))
    {
        memcpy((void *)dst, (void *)p_lssm->image_header, p_lssm->header_length);
    }
}

/*!
 * @brief Copy image header from caller's buffer out to shared memory buffer
 *
 * If caller's image header exceeds the capacity of the shared memory image
 * buffer area, then not bytes are copied, and the current image header contents
 * (if any) are effectively deleted by setting the image header length to zero.
 *
 * @param p_shared_mem Pointer to opaque LSSM shared memory area
 * @param src          Pointer to caller's buffer, containing image header
 * @param src_size     Size (bytes) of image header
 */
void lssm_copyout_image_header(void *p_shared_mem, uint8_t *src, uint32_t src_size)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return;
    }

    if (src_size > sizeof(p_lssm->image_header))
    {
	syslog(LOG_ERR, "%s: src_size(%u) > %u", __PRETTY_FUNCTION__, src_size, sizeof(p_lssm->image_header));
        return;
    }

    if (src_size > 0)
        memcpy((void *)p_lssm->image_header, (void *)src, src_size);
    p_lssm->header_length = src_size;
}

/*!
 * @brief Mark a buffer as full and ready to be read
 *
 * When the server has finished writing to the buffer, the next state
 * of the buffer is BFSTATE_FULL.
 *
 * @param p_shared_mem  Pointer to shared memory.
 * @param bufid         ID of the buffer whose address is to be returned.
 */
void lssm_set_write_done_l(void *p_shared_mem, int bufid)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return;
    }

    if ((bufid < 0) || (bufid >= NUM_BUFS))
    {
        syslog(LOG_ERR, "%s: bad bufid=%d", __FUNCTION__, bufid);
    }
    else
    {
        p_lssm->imagebuf[bufid].state = BFSTATE_FULL;
    }
}

/*!
 * @brief Get the oldest full frame
 *
 * When doing download, this gets the next frame in the download sequence.
 * The buffer is marked as BFSTATE_READING. The buffer must be released
 * when the read is complete.
 *
 * @param p_shared_mem      Pointer to shared memory.
 *
 * @returns buffer ID >= 0 of the oldest full buffer, negative for error
 *
 * @see lssm_release_buffer_l
 */
int lssm_get_oldest_l(void *p_shared_mem)
{
    lssm_t  *p_lssm = (lssm_t *)p_shared_mem;
    int      oldest;
    uint32_t oldest_sequence;
    int      i;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return -1;
    }

    oldest = -1;
    oldest_sequence = 0;
    for (i = 0; i < NUM_BUFS; ++i)
    {
        if (p_lssm->imagebuf[i].state == BFSTATE_FULL)
        {
            if ((oldest == -1) ||
                (oldest_sequence > p_lssm->imagebuf[i].sequence))
            {
                oldest = i;
                oldest_sequence = p_lssm->imagebuf[i].sequence;
            }
        }
    }
    /* If we didn't find any full buffers, return that as an error. */
    if (oldest == -1)
    {
//	syslog(LOG_ERR, "%s: No full buffers!", __PRETTY_FUNCTION__);
        return -1;
    }

    /* Otherwise, this is our guy. */
    p_lssm->imagebuf[oldest].state = BFSTATE_READING;

    return oldest;
}

/*!
 * @brief Get the newest full frame
 *
 * When doing playback or singlestep, this gets the newest frame that has been
 * processed for display. Any older buffers that are also BFSTATE_FULL are
 * released as BFSTATE_EMPTY so they can be reused.
 *
 * The acquired buffer is marked as BFSTATE_READING. The buffer must be released
 * when the read is complete.
 *
 * @param p_shared_mem      Pointer to shared memory.
 *
 * @returns buffer ID >= 0 of the oldest full buffer, negative for error
 *
 * @see lssm_release_buffer_l
 */
int lssm_get_newest_l(void *p_shared_mem)
{
    lssm_t  *p_lssm = (lssm_t *)p_shared_mem;
    int      newest;
    uint64_t newest_sequence;
    int      i;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return -1;
    }

    newest = -1;
    newest_sequence = 0;
    for (i = 0; i < NUM_BUFS; ++i)
    {
        if (p_lssm->imagebuf[i].state == BFSTATE_FULL)
        {
            if ((newest == -1) ||
                (newest_sequence < p_lssm->imagebuf[i].sequence))
            {
                newest = i;
                newest_sequence = p_lssm->imagebuf[i].sequence;
            }
        }
    }

    /* If we didn't find any full buffers, return that as an error. */
    if (newest == -1)
    {
//	syslog(LOG_ERR, "%s: No full buffers!", __PRETTY_FUNCTION__);
        return -1;
    }

    /* Otherwise, this is our guy. */
    p_lssm->imagebuf[newest].state = BFSTATE_READING;

    /* Empty any other full buffers. They should all be older than ours. */
    for (i = 0; i < NUM_BUFS; ++i)
    {
        if (p_lssm->imagebuf[i].state == BFSTATE_FULL)
        {
            if (newest_sequence > p_lssm->imagebuf[i].sequence)
                p_lssm->imagebuf[i].state = BFSTATE_EMPTY;
            else
                syslog(LOG_ERR, "%s: found older buffer %llu>%llu!", __FUNCTION__,
                       newest_sequence, p_lssm->imagebuf[i].sequence);
        }
    }

    return newest;
}

/*!
 * @brief Set a buffer to the EMPTY state.
 *
 * Typically used by the reader when it has finished reading
 * the buffer content. Could also be used by the writer if
 * it changes its mind after setting BFSTATE_WRITING and
 * decides not to mark the buffer as full after all, probably
 * because of an error or cancelation.
 *
 * @param p_shared_mem  Server's pointer to shared memory.
 * @param bufid         ID of buffer to release
 */
void lssm_release_buffer_l(void *p_shared_mem, int bufid)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return;
    }

    if ((bufid < 0) || (bufid >= NUM_BUFS))
    {
        syslog(LOG_ERR, "%s: bad bufid=%d", __FUNCTION__, bufid);
    }
    else
    {
        p_lssm->imagebuf[bufid].state = BFSTATE_EMPTY;
    }
}

/*!
 * @brief Initialize a process-sharable mutex
 *
 * @param   pm      Pointer to mutex to initialize
 *
 * @return zero for success, non-zero for failure
 */
static int setup_shared_mutex(pthread_mutex_t *pm)
{
    pthread_mutexattr_t mutex_attr;
    int rc;

    // Sanity check.
    if (pm == NULL)
    {
	syslog(LOG_ERR, "%s: mutex pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    /* Initialize mutex attribute variable. */
    rc = pthread_mutexattr_init(&mutex_attr);
    if (rc != 0)
    {
	syslog(LOG_ERR, "%s: pthread_mutexattr_init() failed, rc=%d!", __PRETTY_FUNCTION__, rc);
        return rc;
    }

    // We do not want recursive for mutex used with condition variable.
    rc = pthread_mutexattr_settype(&mutex_attr, PTHREAD_MUTEX_NORMAL);
    if (rc != 0)
    {
	syslog(LOG_ERR, "%s: pthread_mutexattr_settype() failed, rc=%d!", __PRETTY_FUNCTION__, rc);
        goto bail;
    }

// ROBUST option is not available the way we are compiled.
//  rc = pthread_mutexattr_setrobust( &mutex_attr, PTHREAD_MUTEX_ROBUST );
//  if ( rc < 0 )
//      goto bail;

    rc = pthread_mutexattr_setpshared(&mutex_attr, PTHREAD_PROCESS_SHARED);
    if (rc != 0)
    {
	syslog(LOG_ERR, "%s: pthread_mutexattr_setpshared() failed, rc=%d!", __PRETTY_FUNCTION__, rc);
        goto bail;
    }

    // Create the shared memory mutex in our shared memory region.
    // We will use this same mutex for the server wakeup condition.
    rc = pthread_mutex_init(pm, &mutex_attr);
    if (rc != 0)
    {
	syslog(LOG_ERR, "%s: pthread_mutex_init() failed, rc=%d!", __PRETTY_FUNCTION__, rc);
        goto bail;
    }

    rc = 0;

bail:
    pthread_mutexattr_destroy(&mutex_attr);
    return rc;
}


/*!
 * @brief Initialize a process-sharable condition variable
 *
 * @param   pc      Pointer to condition variable to initialize
 *
 * @return zero for success, non-zero for failure
 */
static int setup_shared_cond(pthread_cond_t *pc)
{
    pthread_condattr_t cond_attr;
    int rc;

    // Sanity check.
    if (pc == NULL)
    {
	syslog(LOG_ERR, "%s: condition pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    rc = pthread_condattr_init(&cond_attr);
    if (rc != 0)
    {
	syslog(LOG_ERR, "%s: pthread_condattr_init() failed, rc=%d!", __PRETTY_FUNCTION__, rc);
        return rc;
    }

    /* Set up attributes for condition variable. */
    rc = pthread_condattr_setpshared(&cond_attr, PTHREAD_PROCESS_SHARED);
    if (rc != 0)
    {
	syslog(LOG_ERR, "%s: pthread_condattr_setpshared() failed, rc=%d!", __PRETTY_FUNCTION__, rc);
        goto bail;
    }

    /* Create server's condition variable in our shared memory region. */
    rc = pthread_cond_init(pc, &cond_attr);
    if (rc != 0)
    {
	syslog(LOG_ERR, "%s: pthread_cond_init() failed, rc=%d!", __PRETTY_FUNCTION__, rc);
        goto bail;
    }

    rc = 0;

bail:
    pthread_condattr_destroy(&cond_attr);
    return rc;
}

/*!
 * @brief Cleanup before Camera Server exit.
 *
 * @param p_shared_mem    Shared memory pointer as seen from server.
 */
void lssm_server_cleanup(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return;
    }

    (void)pthread_cond_destroy(&p_lssm->server_wakeup);
    (void)pthread_cond_destroy(&p_lssm->client_wakeup);
    (void)pthread_mutex_destroy(&p_lssm->stream_lock);

    (void)munmap(p_shared_mem, LOCALSTREAM_SHM_SIZE);
    (void)shm_unlink(LOCALSTREAM_SHM_NAME);
}

/*!
 * @brief Cleanup when switching modes.
 *
 * Make sure there are no owned buffers.
 *
 * @param p_shared_mem    Shared memory pointer as seen from server.
 *
 * @return zero for success, non-zero for failure
 */
int lssm_server_reset(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int     i;
    int     rc;
    struct timespec ts;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    // Wait long enough...but not too long.
    clock_gettime(CLOCK_REALTIME, &ts);
    ts.tv_sec += 2;

    // Lock access.
    rc = lssm_timedlock(p_shared_mem, &ts);
    if (rc != 0)
        syslog(LOG_ERR, "%s: failed to lock (%d) %s", __FUNCTION__, rc, strerror(rc));

    // Return any non-empty buffers.
    for (i = 0; i < NUM_BUFS; ++i)
    {
        if (p_lssm->imagebuf[i].state != BFSTATE_EMPTY)
        {
            syslog(LOG_ERR, "%s: bad cleanup on previous operation, buffer #%d state=%u",
                   __FUNCTION__, i, p_lssm->imagebuf[i].state);
            p_lssm->imagebuf[i].state = BFSTATE_EMPTY;
        }
    }

    p_lssm->server_fault = 0; // If client has not cooperated, force the fault flag clear
    p_lssm->client_fault = 0; // If client has not cooperated, force the fault flag clear

    // Go ahead and unlock in case the error is that we already own the lock.
    if ((rc == 0) || (rc == EDEADLK))
        rc = lssm_unlock(p_shared_mem);

    return rc;
}

/*!
 * @brief Initialize the Local Stream Shared Memory variables.
 *
 * @param p_lssm    Pointer to state structure in shared memory
 *
 * @returns 0 for success, non-zero for error.
 */
static int lssm_memory_init(
    lssm_t   *p_lssm,
    uint64_t *buf_addrs,
    uint32_t  buf_size,
    unsigned  buf_count)
{
    int rc;
    int i;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        rc = EINVAL;
        goto bail;
    }

    // Clear out our shared memory structure.
    memset((void *)p_lssm, 0, sizeof(p_lssm));

    rc = setup_shared_mutex(&p_lssm->stream_lock);
    if (rc != 0)
        goto bail;

    rc = setup_shared_cond(&p_lssm->server_wakeup);
    if (rc != 0)
    {
        (void)pthread_mutex_destroy(&p_lssm->stream_lock);
        goto bail;
    }

    rc = setup_shared_cond(&p_lssm->client_wakeup);
    if (rc != 0)
    {
        (void)pthread_cond_destroy(&p_lssm->server_wakeup);
        (void)pthread_mutex_destroy(&p_lssm->stream_lock);
        goto bail;
    }

    // Copy the P1 ping-pong image buffer information we received.
    p_lssm->buf_size = buf_size;
    for (i = 0; i < NUM_BUFS; ++i)
    {
        p_lssm->imagebuf[i].state = BFSTATE_EMPTY;
        p_lssm->imagebuf[i].im_addr = buf_addrs[i];
    }
    p_lssm->last_sequence = 0;
    p_lssm->header_length = 0;
    p_lssm->server_fault = 0;
    p_lssm->client_fault = 0;

    rc = 0;

bail:
    // Log any errors.
    if (rc != 0)
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));

    return rc;
}

/*!
 * @brief Initialize the Local Stream Shared Memory for the server.
 *
 * The Camera Server runs before the web socket server starts and
 * does this server initialization when it first starts up so it
 * is ready for the WSS to connect to it.
 *
 * @returns non-zero for error, 0 for success.
 */
int lssm_server_init(
    void      **p_shared_mem,
    uint64_t  *buf_addrs,
    uint32_t  buf_size,
    unsigned  buf_count)
{
    int   fd;
    int   rc;

    if (sizeof(lssm_t) > LOCALSTREAM_SHM_SIZE)
    {
        syslog(LOG_ERR, "%s: shared memory size too small %u > %u", __FUNCTION__,
               sizeof(lssm_t), LOCALSTREAM_SHM_SIZE);
        return -1;
    }

    // These should match. We just use separate definitions because the web socket server
    // will link with this code and doesn't know about partition manager.
    if (NUM_BUFS > buf_count)
    {
        syslog(LOG_ERR, "%s: more shared memory buffers (%u) than in partition manager (%u)",
               __FUNCTION__, NUM_BUFS, buf_count);
        return -1;
    }

    // Make sure we have buffers.
    if (buf_addrs == NULL)
    {
        syslog(LOG_ERR, "%s: No buffers provided!", __FUNCTION__);
        return -1;
    }

    // Open up our shared memory anchor.
    fd = shm_open(LOCALSTREAM_SHM_NAME, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
    if (fd < 0)
    {
        syslog(LOG_ERR, "%s: shm_open failed (%d) %s", __FUNCTION__, errno,
               strerror(errno));
        return -1;
    }

    // Make sure we set the size of the shared memory area.
    if (ftruncate(fd, LOCALSTREAM_SHM_SIZE) < 0)
    {
        syslog(LOG_ERR, "%s: ftruncate failed (%d) %s", __FUNCTION__, errno,
               strerror(errno));
        close(fd);
        return -1;
    }

    // Map our shared memory anchor into our memory space.
    *p_shared_mem = mmap(0, LOCALSTREAM_SHM_SIZE, PROT_READ | PROT_WRITE,
                         MAP_SHARED, fd, 0);
    if (*p_shared_mem == MAP_FAILED)
    {
        syslog(LOG_ERR, "%s: mmap failed (%d) %s", __FUNCTION__, errno,
               strerror(errno));
        close(fd);
        return -1;
    }

    /* Don't need this any more once the mmap is done. */
    close(fd);

    // Initialize the shared memory control structures.
    rc = lssm_memory_init((lssm_t *)*p_shared_mem, buf_addrs, buf_size, buf_count);
    if (rc != 0)
    {
        (void)munmap(*p_shared_mem, LOCALSTREAM_SHM_SIZE);
        *p_shared_mem = NULL;

        (void)shm_unlink(LOCALSTREAM_SHM_NAME);
    }

    return rc;
}


/*!
 * @brief Wake up the client using the condition variable in shared memory.
 *
 * This needs to be a separate function from the lock/unlock to support proper
 * mutual exclusion when signalling:
 *
 *   lock
 *   set up memory content, e.g. get buffer, set mode or frame number
 *   signal condition
 *   unlock
 *
 * @param p_shared_mem Client's pointer to shared memory.
 *
 * @returns non-zero for error, 0 for success.
 */
int lssm_client_signal_l(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    rc = pthread_cond_signal(&p_lssm->client_wakeup);
    if (rc != 0)
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));

    return rc;
}

/*!
 * @brief Client call to wait for event or timeout
 *
 * @param p_shared_mem  Server's pointer to shared memory.
 * @param timeout_usecs Timeout in microseconds
 *
 * @returns non-zero for error, 0 for success.
 */
int lssm_client_wait_l(void *p_shared_mem, uint64_t timeout_usecs)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    struct timespec ts;
    int             rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    if (timeout_usecs != 0)
    {
        clock_gettime(CLOCK_REALTIME, &ts);
        ts.tv_nsec += ((timeout_usecs % 1000000) * 1000);
        while (ts.tv_nsec >= 1000 * 1000 * 1000)
        {
            ts.tv_nsec -= 1000 * 1000 * 1000;
            ts.tv_sec += 1;
        }
        ts.tv_sec += (timeout_usecs / 1000000);

        rc = pthread_cond_timedwait(&p_lssm->client_wakeup, &p_lssm->stream_lock, &ts);
    }
    else
    {
        rc = pthread_cond_wait(&p_lssm->client_wakeup, &p_lssm->stream_lock);
    }

    if ((rc != 0) && (rc != ETIMEDOUT))
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));

    return rc;
}

int lssm_client_clear_fault_l(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int i;
    int rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    // Return any buffers that we own.
    for (i = 0; i < NUM_BUFS; ++i)
    {
        if ((p_lssm->imagebuf[i].state == BFSTATE_FULL) ||
            (p_lssm->imagebuf[i].state == BFSTATE_READING))
        {
            p_lssm->imagebuf[i].state = BFSTATE_EMPTY;
        }
    }

    // Clear the fault indication and wake up the server.
    p_lssm->server_fault = 0;
    rc = pthread_cond_signal(&p_lssm->server_wakeup);
    if (rc != 0)
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));

    return rc;
}

int lssm_client_set_fault_l(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int rc;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return EINVAL;
    }

    syslog(LOG_ERR, "%s: Send LSSM Client fault indication to LSSM server", __FUNCTION__);
    p_lssm->client_fault = 1;
    rc = pthread_cond_signal(&p_lssm->server_wakeup);
    if (rc != 0)
        syslog(LOG_ERR, "%s: failed (%d) %s", __FUNCTION__, errno, strerror(errno));

    return rc;
}

/*!
 * @brief Initialize the Local Stream Shared Memory for the client (WSS)
 *
 * The Camera Server runs before the web socket daemon starts and does
 * server initialization when it first starts up.
 * It should always be ready for the WSS (web socket server) to connect,
 * the client should find the share existing already.
 *
 * @returns non-zero for error, 0 for success.
 */
int lssm_client_init(void **p_shared_mem)
{
    int   fd;

    fd = shm_open(LOCALSTREAM_SHM_NAME, O_RDWR, S_IRUSR | S_IWUSR);
    if (fd < 0)
    {
        syslog(LOG_ERR, "%s: shm_open failed (%d) %s", __FUNCTION__, errno,
               strerror(errno));
        return -1;
    }

    *p_shared_mem = mmap(0, LOCALSTREAM_SHM_SIZE, PROT_READ | PROT_WRITE,
                         MAP_SHARED, fd, 0);
    if (*p_shared_mem == MAP_FAILED)
    {
        syslog(LOG_ERR, "%s: mmap failed (%d) %s", __FUNCTION__, errno,
               strerror(errno));
        close(fd);
        return -1;
    }

    /* Don't need this any more once the mmap is done. */
    close(fd);

    return 0;
}

/*!
 * @brief Cleanup when client is shutting down a session; release all buffers
 *
 * @param p_shared_mem    Shared memory pointer as seen from server.
 * @return The number of buffers reelased
 */
int lssm_client_cleanup_l(void *p_shared_mem)
{
    lssm_t *p_lssm = (lssm_t *)p_shared_mem;
    int     i;
    int     rc = 0;

    // Sanity check.
    if (p_lssm == NULL)
    {
	syslog(LOG_ERR, "%s: shared memory pointer is null!", __PRETTY_FUNCTION__);
        return 0;
    }

    // Return any buffers that we own.
    for (i = 0; i < NUM_BUFS; ++i)
    {
        if (p_lssm->imagebuf[i].state != BFSTATE_EMPTY)
        {
            p_lssm->imagebuf[i].state = BFSTATE_EMPTY;
            rc++;
        }
    }
    return rc;
}

#include <sys/syscall.h>
static pid_t lssm_gettid()
{
    uint32_t    tid;
    tid = (uint32_t)syscall(SYS_gettid);

    return tid;
}

